#include <stdlib.h>
#include "share_string.h"

// A small C library for avoiding duplicate strings, which saves space.
// It is a very bad idea to use this if code might mutate a string's contents.
// Freeing a string-cache recursively frees all strings in it!
// (For you homework, do not bother because you need the strings until the
//  program exits.)

// This library is not particularly efficient; one should really use hashing
// or a trie or something, but this "slow" way may be easier to understand.

// It uses a classic "resizing array" technique.

struct WordCache {
  int arr_len;
  int next_slot;
  char ** arr;
};

words_t new_word_cache() {
  words_t ans = (words_t)malloc(sizeof(struct WordCache));
  ans->arr_len = 10;
  ans->next_slot = 0;
  ans->arr = (char**)malloc(10 * sizeof(char*));
  return ans;
}

void free_cache(words_t w) {
  int i;
  for(i=0; i < w->next_slot; ++i)
    free((w->arr)[i]);
  free(w);
}

void resize_cache(words_t w) {
  int new_len = w->arr_len * 2;
  char ** new_arr = (char**)malloc(new_len * sizeof(char*));
  int i;
  for(i=0; i < w->arr_len; ++i)
    new_arr[i] = (w->arr)[i];
  free(w->arr);
  w->arr_len = new_len;
  w->arr = new_arr;
}

// n is the number of characters to consider in the input, which is
// convenient so you can pass a "string" that is _not_ terminated with '\0'.
// What you get back will be a malloc'ed array that does end with '\0'.
char * find_word(words_t w, char * s, int n) {
  int i;
  for(i=0; i < w->next_slot; ++i) {
    char * s2 = (w->arr)[i];
    int j;
    // stop early if a character differs or s2 is too short
    for(j=0; j < n; ++j) {
      if(s2[j] != s[j] || s2[j] == '\0')
	break;
    }
    // we found a match if we did not stop early and s2 is not any longer
    if(j==n && s2[n] == '\0')
      return s2;
  }
  // no match found, add a copy of the string to the cache
  char * new_word = (char*)malloc((n+1)*sizeof(char));
  for(i=0; i < n; ++i)
    new_word[i] = s[i];
  new_word[n] = '\0';
  if(w->next_slot == w->arr_len)
    resize_cache(w);
  (w->arr)[w->next_slot] = new_word;
  w->next_slot += 1;
  return new_word;
}

